#!/usr/bin/python
"""
A user interface class implemented using pygame. A plot are and a group of buttons placed
below are implemented. Buttons are added along with a callback function that is not a
member of the class.

"""

import pygame, sys
from pygame.locals import *
from pygrace import *
import phoenix

# global variables
NP = 128						# Number of points in x direction
width = 500						# diplay window dimensions
height = 400

w = None						# Global Pwin object
p = phoenix.phoenix()			# The Phoenix object
delay = 120						# Phoenix ADC conversion time

#parameters required by Pwin for scaling the graphs
xmin = 0
xmax = delay * NP/1000000.0 	# Maximum possible value of time
ymin = -5000.0					# This will be zero for unipolar ADC inputs
ymax = +5000.0

looping = False					# global variable for programs like CRO
trig = 0.0						# used by CRO
phdata = []						# data from phoenix

class Pwin:
	""" A class to manage a window with a plot area and several buttons"""
	global xmin, xmax, ymin, ymax, NP, phdata
	border = 3
	menuh = 30
	plotw = width - border * 2
	ploth = height - menuh - border * 4
	plotrect = pygame.Rect(border, border, plotw+border, ploth+border)
	xscale = plotw / (xmax - xmin)
	yscale = ploth / (ymax - ymin)
	white = (255, 255, 255)
	black = (0, 0, 0)
	green = (0, 100, 0)
	numButtons = 0
	left = 0;					# static used by addButton()
	buttons = []
	bg = None					# pygame Surface object
	scr = None					# pygame display object
	font = None
	points = []					# Array of points required by pygame.draw_lines
	
	def __init__(self, title, width = 500, height = 400):
		pygame.init()
		self.scr = pygame.display.set_mode((width, height))
		pygame.display.set_caption(title)
		self.bg = pygame.Surface(self.scr.get_size())
		self.bg = self.bg.convert()
		self.bg.fill(self.black)
		self.font = pygame.font.Font(None, 24)
		for x in range(NP):					# Is there a better way to create this array
			self.points.append([0,0])  		# this is made global to avoid creating it every time

	def addButton(self, name, func):
		""" Adds a Button with label 'name' and make the binding to call 'func' when clicked
		"""
		text = self.font.render(name , 1, self.white)
		textrect = text.get_rect()
		textrect.bottomleft = (self.border + self.left, height - self.border)
		pygame.draw.rect(self.bg, self.green, textrect, 0)
		self.bg.blit(text, textrect)
		self.buttons.append((name, func, textrect))	
		self.left = self.left + textrect.width + 2 * self.border
		self.numButtons = self.numButtons + 1

	def refresh(self):
		self.scr.blit(self.bg, (0,0))
		pygame.display.flip()

	def clear_plot(self):
		self.bg.fill(self.black, self.plotrect)
	
		
	def draw_lines(self):
		"""	plots lines using the global variable points.
			Also uses xscale, yscale and ymin to scale the coordinates to plot window
		"""

		self.clear_plot()
		i = 0
		while i < NP:
			self.points[i][0] = self.border + phdata[i][0] * self.xscale
			self.points[i][1] = self.border + self.ploth - (phdata[i][1] - ymin) * self.yscale
#			print points[x][0], ' ', points[x][1]
			i = i + 1
	
		pygame.draw.lines(self.bg, self.white, 0, self.points, 1)
		self.refresh()


	def disp_image(img):
		pic = pygame.image.load(img)
		rect = pic.get_rect()
		rect.right = self.plotw		# restrict image to plot area
		rect.bottom = self.ploth
		self.scr.blit(pic, rect)
		pygame.display.flip()

	def	userInput(self):									# Event loop
		""" infinite loop lokking for mouse down events"""
		global looping
		while 1:
			for e in pygame.event.get():
				if e.type == MOUSEBUTTONDOWN:
					x = 0
					cmd = None
					while x < self.numButtons:
						left = self.buttons[x][2][0]
						top = self.buttons[x][2][1]
						right = left + self.buttons[x][2][2]
						bottom = top + self.buttons[x][2][3]
						if left < e.pos[0] < right and top < e.pos[1] < bottom:
							self.buttons[x][1]()		# call the associated function
						x = x +1
			periodic()
#-------------------------------------class end ----------------------------------------			

def	periodic():		
	""" Called continuosly by the event handler loop. Anything that has to be done in a loop
	call be called from here by setting appropriate global flags.
	"""
	global looping
	if looping:
		readADC()
		w.draw_lines()
		
def	readADC():
	"""
	Reads NP points from selected ADC channel and puts in the global variable "phdata"
	waits for reaching the specified trigger voltage before reading the block
	"""
	global	trig, NP, delay, points, phdata
	for k in range(NP):  
		val = p.minus5000_to_5000()
		if( val > trig and val < trig + 200):
			break
	phdata = p.read_block(NP, delay, 1)
				
def	start():
	global looping
	looping = True

def	stop():
	global looping
	looping = False

def	single():
	readADC()
	w.draw_lines()
	
def analyse():
	global NP, phdata
	pg = grace()
	x = range(NP)
	y = range(NP)
	i = 0
	while i < NP:				# data returned by read_block goes to xmgrace
		x[i] = phdata[i][0]		
		y[i] = phdata[i][1]	
		i = i + 1
	pg.plot(x,y)

def save():
	global  NP, phdata
	f = open('cro.dat','w')
	i = 0
	while i < NP:				# data returned by read_block goes to file
		x = phdata[i][0]
		y = phdata[i][1]
		s = str(x) + ' ' + str(y) + '\n'	
		f.write(s)	
		i = i + 1
	f.close()	


def one():
	p.select_adc(0)

def two():
	p.select_adc(1)

def	quit():
	print 'bye bye'
	sys.exit()

p.select_adc(0)
phdata = p.read_block(NP, delay, 1)

w = Pwin("Phoenix CRO Program")
w.addButton("Start", start)
w.addButton("Stop", stop)
w.addButton("Single", single)
w.addButton(" 1 ", one)
w.addButton(" 2 ", two)
w.addButton("Save", save)
w.addButton("Analyse", analyse)
w.addButton("Quit", quit)

w.refresh()			# display everything now
w.userInput()		# The main loop
